/**
 * Created by Connor Weeks on 3/18/2017.
 */


// when true, clicks are tracked with coordinates in user mode
var DEBUG = false;

// center coordinates for the map of Lambeth
var LAMBETH_CENTER_LAT = 51.46021835;
var LAMBETH_CENTER_LNG = -0.121445154447924;

// starting zoom level to show all of Lambeth
var DEFAULT_ZOOM = 13;

// the maximum zoom level in which zones are filled with color
var MAX_ZONE_DISPLAY_ZOOM = 14;

// variable which will hold the mpa object
var map;

/**
 * the object containing both the type and ID of the currently selected object.
 * The type field should be one of TYPE_ZONE or TYPE_LOCATION
 */
var selected = {
    type: null,
    id: null
};

// object obtained from the server representing each parking zone's data
var zoneData;

// object obtained from the server representing each parking location's data
var locationData;

// the string identifying a geoJson parking location
var TYPE_LOCATION = 'location';

// the string identifying a geoJson CPZ
var TYPE_ZONE = 'zone';

// the string identifying a geoJson multipoint for pay and display machines
var TYPE_PD = 'pd';

// whether or not to show pay and display machines on the map
var showPd;

// whther or not the mouse button is currently pressed
var mouseDown = false;

// whether or not the user is currently parking; this is used in initMap but only changed in index.js
var currentlyParking = false;

// whether or not to keep highlighting zones when zoomed in
var highlightWhenZoomed = false;


/**
 * Returns an object of the data about zone and location from the given ID.
 * If a zone's ID and type (TYPE_ZONE) are given, the returned values will
 * have fields valued 'None' for location data.  If a location ID and type
 * (TYPE_LOCATION) is give, all fields will be valid.
 * @param id The ID of the object to obtain the data of.
 * @param type the type of the object the iD represents (either TYPE_ZONE or
 * TYPE_LOCATION).
 * @returns {*}
 */
function getData(id, type) {
    // define default values for location fields
    var locationName = 'None';
    var locationId = 'None';
    var zoneId;
    var capacity;
    var currentOccupancy;

    // determine the zone id either directly or from indexing a location's fields
    if (type == TYPE_ZONE && zoneData[id]) {
        zoneId = id;
        capacity = zoneData[id].capacity;
        currentOccupancy = zoneData[id].currentOccupancy;
    } else if (type == TYPE_LOCATION && locationData[id] && zoneData[locationData[id].zoneId]) {
        locationId = id;
        var location = locationData[locationId];
        locationName = location.name + ' (' + locationId + ')';
        capacity = location.capacity;
        currentOccupancy = location.currentOccupancy;
        zoneId = location.zoneId;
    } else {
        return null;
    }

    // return data about the zone and location
    var zone = zoneData[zoneId];
    return {
        zoneId: zoneId,
        zoneName: zone.name + ' (' + zoneId.toUpperCase() + ')',
        band1Price: zone.band1Price,
        band6Price: zone.band6Price,
        occupancyMultiplier: zone.occupancyMultiplier,
        locationId: locationId,
        locationName: locationName,
        capacity: capacity,
        currentOccupancy: currentOccupancy,
        fractionOccupied: Math.min(currentOccupancy / capacity, 1)
    }
}


/**
 * Sets the selected object based on a given ID and type.
 * If the given ID does not exist, no action is taken.
 * @param id The ID of the object to select.
 * @param type The type of the object to selected; must
 * be either TYPE_ZONE or TYPE_LOCATION.
 */
function setSelected(id, type) {
    // check if the given ID and type are valid
    if ((type == TYPE_ZONE && zoneData[id]) || (type == TYPE_LOCATION && locationData[id])) {
        if (id != selected.id) {
            // if we are selecting a new object, update the map
            selected.id = id;
            selected.type = type;
            displayData(id, type);

        } else {
            // if the user deselected the current object, set the fields of selected to null
            selected.id = null;
            selected.type = null;
        }

        // reset style to account for changes
        if (map) {
            map.data.setStyle(setStyle);
        }
    }
}

function unselect() {
    selected.id = null;
    selected.type = null;
    // reset style to account for changes
    if (map) {
        map.data.setStyle(setStyle);
    }
}


function refreshZoneList(includeNoneOption) {
    var zoneNames = [];
    var zonesSelect = $('#parking-zones');
    if (!zonesSelect.length) {
        return;
    }
    var currentlySelected = zonesSelect.find(':selected').val();

    zonesSelect.empty();

    if (includeNoneOption) {
        zonesSelect.append('<option id="none"></option>');
    }

    for(var zoneId in zoneData) {
        if (zoneData.hasOwnProperty(zoneId)) {
            zoneNames.push({
                zoneId: zoneId,
                name: zoneData[zoneId].name + ' (' + zoneId.toUpperCase() + ')'
            });
        }
    }
    zoneNames.sort(function (firstZone, secondZone) {
        var names = [firstZone.name, secondZone.name];
        names.sort();
        if (names[0] == firstZone.name) {
            return -1;
        }
        return 1;
    });
    for (var i = 0; i < zoneNames.length; i++) {
        zonesSelect.append($('<option>', {
                value: zoneNames[i].zoneId,
                text: zoneNames[i].name
            })
        )
    }
    // reset the currently selected value to its previous state
    zonesSelect.val(currentlySelected);
}

function refreshLocationList(includeNoneOption) {
    var locationNames = [];
    var locationsSelect = $('#parking-locations');
    if (!locationsSelect.length) {
        return;
    }
    var currentlySelected = locationsSelect.find(':selected').val();

    locationsSelect.empty();

    if (includeNoneOption) {
        locationsSelect.append('<option id="none"></option>');
    }

    for(var locationId in locationData) {
        if (locationData.hasOwnProperty(locationId)) {
            locationNames.push({
                locationId: locationId,
                name: locationId + ' - ' + locationData[locationId].name
            });
        }
    }
    locationNames.sort(function (firstLocation, secondLocation) {
        return secondLocation - firstLocation;
    });
    for (var i = 0; i < locationNames.length; i++) {
        locationsSelect.append($('<option>', {
                value: locationNames[i].locationId,
                text: locationNames[i].name
            })
        )
    }
    // reset the currently selected value to its previous state
    locationsSelect.val(currentlySelected);
}



/**
 * Returns a color between green and red,
 * depending on the input number
 * @param value A number between 0 and 1:
 * 0 is completely green and 1 is
 * completely red
 * @returns {string} Color string
 */
function getColor (value){
    // value from 0 to 1
    var hue = ((1 - value) * 120).toString(10);
    return ['hsl(', hue, ',100%,50%)'].join('');
}

/**
 * Returns a linear interpolation from 4 given base
 * values and one intermediate value.
 * @param x1 The first independent variable value.
 * @param x2 The second independent variable value.
 * @param y1 The first dependent variable value.
 * @param y2 The second dependent variable value.
 * @param x The intermediate independent variable value,
 * for which we wish to find a dependent variable value.
 * @returns The result of the linear interpolation.
 */
function interpolate(x1, x2, y1, y2, x) {
    var ratio = (x - x1) / (x2 - x1);
    return y1 + ratio * (y2 - y1);

}


function printLocation(feature) {
    var locationId = feature.getProperty('locationId');
    var name = feature.getProperty('name');
    var zoneId = feature.getProperty('zoneId');
    var min = 10;
    var max = 30;
    var capacity = Math.floor(Math.random() * (max - min + 1)) + min;

    console.log("" + locationId + ": {\n    name: '" + name + "',\n    zoneId: '" + zoneId + "',\n    capacity: " + capacity + ",\n    currentOccupancy: 0\n},");
}

/**
 * This sets the style of the map, given a feature.
 * @param feature the feature to apply styling to.
 */
function setStyle(feature) {
    var defaultStyle = {
        fillColor: 'black',
        fillOpacity: 0.02,
        strokeWeight: 1,
        clickable: DEBUG
    };

    var opacity;
    var strokeWeight = 1;
    var color;
    var clickable = true;

    try {
        var type = feature.getProperty('type');
        var zoom = map.getZoom();
        if (type == TYPE_ZONE) {

            var zoneId = feature.getProperty('zoneId');
            var ratio = zoneData[zoneId].currentOccupancy / zoneData[zoneId].capacity;
            color = getColor(Math.min(ratio, 1));
            if (zoneId == selected.id) {
                opacity = 0.4;
                strokeWeight = 2;
            }
            clickable = zoom <= MAX_ZONE_DISPLAY_ZOOM;
            if (zoom > MAX_ZONE_DISPLAY_ZOOM && !highlightWhenZoomed) {
                // don't highlight the CPZ's unless we are zoomed out enough or the user has requested such
                opacity = 0;
            }
        }
        else if (type == TYPE_LOCATION) {
            //printLocation(feature);
            var locationId = feature.getProperty('locationId');
            var strokeWeight = 4;
            var ratio = locationData[locationId].currentOccupancy / locationData[locationId].capacity;

            color = getColor(Math.min(ratio, 1));
            var strokeOpacity = 0.6;

            if (locationId == selected.id) {
                strokeOpacity = 0.9;
                strokeWeight = 6;
            }

            if (zoom <= MAX_ZONE_DISPLAY_ZOOM) {
                // don't show the individual parking locations unless we are zoomed in enough
                strokeOpacity = 0;
            }
            return {
                strokeColor: color,
                strokeWeight: strokeWeight,
                strokeOpacity: strokeOpacity,
                clickable: true,
                visible: zoom > MAX_ZONE_DISPLAY_ZOOM,
                zIndex: 100
            }
        } else if (type == TYPE_PD) {
            return {
                visible: zoom > MAX_ZONE_DISPLAY_ZOOM && showPd
            };
        } else {
            return defaultStyle;
        }
    } catch (error) {
        return defaultStyle;
    }

    return {
        fillColor: color,
        fillOpacity: opacity,
        strokeWeight: strokeWeight,
        clickable: clickable
    }
}




/**
 * Creates the map and adds event listeners for functionality.
 * This function is called as a callback after receiving the API.
 */
function initMap() {
    map = new google.maps.Map(document.getElementById('map'), {
        center: {lat: LAMBETH_CENTER_LAT, lng: LAMBETH_CENTER_LNG},
        zoom: DEFAULT_ZOOM
    });

    map.data.loadGeoJson('/public/assets/lambeth-zones.json');
    map.data.loadGeoJson('/public/assets/lambeth-pd-locations.json');
    map.data.loadGeoJson('/public/assets/lambeth-locations-waterloo.json');
    map.data.loadGeoJson('/public/assets/lambeth-locations-kennington.json');

    // style the zones and locations
    map.data.setStyle(setStyle);

    // track coordinates if in debug mode (this is used to create CPZs)
    map.data.addListener('click', function(event) {
        if (!DEBUG) {
            return;
        }
        var latitude = event.latLng.lat();
        var longitude = event.latLng.lng();
        console.log('[' + longitude + ', ' + latitude + '],');

        radius = new google.maps.Circle({
            map: map,
            radius: 10,
            center: event.latLng,
            fillColor: '#777',
            fillOpacity: 0.1,
            strokeColor: '#AA0000',
            strokeOpacity: 0.8,
            strokeWeight: 1,
            draggable: false,    // Dragable
            editable: false      // Resizable
        });

    });

    // show zone data when moused over (only if no zone is selected)
    map.data.addListener('mouseover', function (event) {
        // return if a zone or location is currently selected
        if (selected.id) {
            return;
        }
        var type = event.feature.getProperty('type');
        if (type == TYPE_ZONE) {
            // identify te zone we are hovering over
            var zoneId = event.feature.getProperty('zoneId');

            // display the data for that zone
            displayData(zoneId, TYPE_ZONE);
        } else if (type == TYPE_LOCATION) {
            // identify the location we are hovering over
            var locationId = event.feature.getProperty('locationId');

            // display the data for that location
            displayData(locationId, TYPE_LOCATION);
        }
    });

    // refresh when zoomed in
    map.addListener('zoom_changed', function () {
        map.data.setStyle(setStyle);
    });

    // reset the data display when the mouse leaves a zone
    map.data.addListener('mouseout', function () {
        if (!selected.id && !mouseDown) {
            displayData();
        }
    });

    // select a zone or location when the user clicks on it
    map.data.addListener('click', function(event) {
        if (currentlyParking) {
            // don't allow the user t change the selected object when parking
            return;
        }
        var type = event.feature.getProperty('type');
        if (type == TYPE_ZONE) {
            var zoneId = event.feature.getProperty('zoneId');
            setSelected(zoneId, TYPE_ZONE);
        } else if (type == TYPE_LOCATION) {
            var locationId = event.feature.getProperty('locationId');
            setSelected(locationId, TYPE_LOCATION);
        }
    });

    // unselect the currently selected object
    map.addListener('click', function () {
        if (!currentlyParking) {
            unselect();
            displayData();
        }
    });

    // call the file-specific setup function if it exists
    if (typeof setupMap === 'function') {
        setupMap()
    }
}


/**
 * Adds the event(s) necessary to respond to a server update.
 * @param socket The socket object used to connect to the server.
 */
function setupUpdateEvent(socket) {
    socket.on('server-update', function (data) {
        // replace the current parking data with the updated information
        zoneData = data.zoneData;
        locationData = data.locationData;

        // refresh the map
        if (map) {
            map.data.setStyle(setStyle);
        }
        refreshLocationList(true);
        refreshZoneList();
        // refresh currently displayed information
        if (selected.id) {
            displayData(selected.id, selected.type);
        }
    });

    // request an initial set of parking data
    socket.emit('get-data');
}